#
# jv:vdb - Visual / VM Debugger
# Copyright (C) 2015-now Jan Vrany
#
# This software is licensed under 'Creative Commons Attribution-NonCommercial 4.0 International License'
#
# You may find a full license text in LICENSE.txt or at http://creativecommons.org/licenses/by-nc/4.0/
#

import argparse
import copy
import itertools
import sys

import gdb


class Error(Exception):
    pass


_VarEvaluators = []


def register_evaluator(callable, replace=True):
    """
    Register `callable` in list of evaluators which are tried when creating
    -vdb-var-create. To evaluate and expression, `callable` is invoked passing
    expression string and must return either `gdb.Value` or `None`.
    """
    _VarEvaluators.append(callable)


def unregister_evaluator(callable):
    """
    Unregister previously registered evaluator.
    """
    if callable in _VarEvaluators:
        _VarEvaluators.remove(callable)


def unregister_evaluators():
    """
    Unregister all evaluators. To be used mainly in tests
    """
    _VarEvaluators = []


def parse_and_eval(expression):
    """
    Parse and evaluate expression and return a gdb.Value() as follows:

      * First, use builtin `gdb.parse_and_eval()`. If builtin succeeds,
        return resulting value.
      * Second, check if `expressions` matches any argument or local
        as returned by frame decorator for current frame, if it does,
        return it.
      * Finally, try to use any registered custom evaluator (see
        `register_evaluator()`. If first non-None value is returned.
      * If none of the above yields a value, an exception is thrown, just
        like `gdb.parse_and_eval()` does if it cannot parse/evaluate given
        expression.
    """

    # First, try builtin gdb.parse_and_eval:
    try:
        return gdb.parse_and_eval(expression)
    except:
        pass

    # Second, try frame decorator frame_args() / frame_locals():
    decorator = next(gdb.frames.execute_frame_filters(gdb.selected_frame(), 0, 0))
    for var in itertools.chain(decorator.frame_args(), decorator.frame_locals()):
        if var.sym == expression:
            return var.value()

    # Firnally, try custom registered expression evaluators:
    for evaluate in _VarEvaluators:
        value = evaluate(expression)
        if value != None:
            return value

    # Nothing works, give up.
    raise Error("Cannot evaluate expression: %s" % expression)

def contents_visualizer(value):
    """
    Find and return the "contents" visualizer for a gdb.Value.
    If value's `gdb.default_visualizer` defines a property
    `contents_visualizer`, property value is used as a factory
    function to create contents visualizer. Otherwise, a
    `gdb.default_visualizer()` is returned.

    This is to be used by UI to implement "Inspect" on
    variables.
    """
    default_visualizer = gdb.default_visualizer(value)
    if hasattr(default_visualizer, "contents_visualizer"):
        return default_visualizer.contents_visualizer(value)
    else:
        return default_visualizer


def set_prompt(prompt="vdb"):
    """
    Set prompt for both, GDBCLI and Python CLI
    """
    gdb.prompt_hook = lambda x: "%s > " % prompt
    sys.ps1 = "%s pi > " % prompt
    sys.ps2 = "%s    > " % (" " * len(prompt))


autodebug = gdb.Parameter('vdb-autodebug', gdb.COMMAND_DATA, gdb.PARAM_AUTO_BOOLEAN)
autodebug.value = False

# Install automatic debugging hook
def __autodebug_excepthook(typ, value, tb):
    if autodebug.value == False: # hasattr(sys, 'ps1') or not sys.stderr.isatty():
        sys.__excepthook__(typ, value, tb)
        return
    if autodebug.value == None: # set to 'auto'
        pass

    import traceback, pdb
    if tb is None:
        print("EXECEPTION IN PYTHON CODE:")
        traceback.print_last()
    else:
        print("EXECEPTION IN PYTHON CODE, ENTERING DEBUGGER:")
        traceback.print_exception(typ, value, tb)
        # ...then start the debugger in post-mortem mode.
        # pdb.pm() # deprecated
        pdb.post_mortem(tb) # more "modern"

sys.excepthook = __autodebug_excepthook
