#
# jv:vdb - Visual / VM Debugger
# Copyright (C) 2015-now Jan Vrany
#
# This software is licensed under 'Creative Commons Attribution-NonCommercial 4.0 International License'
#
# You may find a full license text in LICENSE.txt or at http://creativecommons.org/licenses/by-nc/4.0/
#
"""
Module vdb.mi contains support for VDB UI. It is not needed when using
standalone GDB.
"""

import argparse
import copy
import itertools
import sys

import gdb

from vdb import Error, parse_and_eval


def micommand(name):
    def _command(func):
        class _Command(gdb.MICommand):
            def invoke(*args, **kwargs):
                return { 'result' : func(*args[1:], **kwargs) }

            def __call__(self, *args, **kwargs):
                return self.invoke(*args, **kwargs)

        return _Command(name)

    return _command


class _VarObjRegistry(object):
    def __init__(self):
        self._seqno = 0
        self._varobjs = {}

    def __iter__(self):
        return iter(self._varobjs)

    def __getitem__(self, name):
        return self._varobjs[name]

    def __delitem__(self, name):
        del self._varobjs[name]

    def generate_name(self):
        """
        Generate and return a new, unique variable name.
        """
        name = "vdbvar" + str(self._seqno)
        self._seqno = self._seqno + 1
        return name

    def validate_name(self, name):
        """
        Check if given `name` is not yet used (and therefore can
        be used for new variable). Throws an error if the name
        is already in use, otherwise no-op.
        """
        if name in self._varobjs:
            raise Error("Name not unique: %s" % name)

    def register(self, varobj):
        self.validate_name(varobj.name)
        self._varobjs[varobj.name] = varobj

    def unregister(self, varobj):
        del self._varobjs[varobj.name]

    def unregister_all(self):
        self._varobjs = {}


_VarObjs = _VarObjRegistry()


class _VarObj(object):
    def __init__(self, name, expr, value, parent=None):
        self._name = name
        self._expr = expr
        self._value = value
        self._parent = parent
        self.set_visualizer(None)

    def __copy__(self):
        return _VarObj(None, self._expr, self._value, self._parent)

    def to_mi(
        self,
        properties=[
            "name",
            "exp",
            "numchild",
            "value",
            "type",
            "thread_id",
            "frozen",
            "has_more",
            "displayhint",
            "dynamic",
        ],
    ):
        """
        Return MI record representing this variable object.
        """
        return {p: getattr(self, p) for p in properties if getattr(self, p) != None}

    def children(self):
        """
        Return an iterable (possibly empty) over child varobjs.
        Children are NOT registered, the caller must take care
        about this.
        """
        if not hasattr(self, "_children"):

            def child(exp, value):
                child_name = self._name + "." + exp
                child = _VarObj(child_name, exp, value, self)
                return child

            if not hasattr(self._visualizer, "children"):
                self._children = []
            else:

                class lazy:
                    def __init__(self, iterator):
                        self._head = []
                        self._rest = iterator

                    def __iter__(self):
                        for each in self._head:
                            yield each
                        for each in self._rest:
                            self._head.append(each)
                            yield each

                self._children = lazy(
                    map(
                        lambda pair: child(pair[0], pair[1]),
                        self._visualizer.children(),
                    )
                )
        return iter(self._children)

    def set_visualizer(self, visualizer):
        """
        Set new visualizer
        """

        self._visualizer = visualizer
        if self._visualizer == None:
            self._visualizer = gdb.default_visualizer(self._value)
        if hasattr(self, "_children"):
            del self._children

    @property
    def parent(self):
        """
        Return parent varobj or None
        """
        return self._parent

    @property
    def name(self):
        """
        The name of the varobj.
        """
        return self._name

    @property
    def numchild(self):
        """
        The number of children of the varobj. This number is not necessarily
        reliable for a dynamic varobj. Instead, you must examine the ‘has_more’
        attribute.
        """
        return 0

    @property
    def value(self):
        """
        The varobj’s scalar value. For a varobj whose type is some sort of
        aggregate (e.g., a struct), or for a dynamic varobj, this value will not
        be interesting.
        """
        if self._visualizer != None:
            return str(self._visualizer.to_string())
        else:
            return str(self._value)

    @property
    def type(self):
        """
        The varobj’s type. This is a string representation of the type, as would
        be printed by the GDB CLI. If ‘print object’ (see set print object) is
        set to on, the actual (derived) type of the object is shown rather than
        the declared one.
        """
        return str(self._value.type)

    @property
    def thread_id(self):
        """
        If a variable object is bound to a specific thread, then this is the
        thread’s global identifier.
        """
        return None

    @property
    def has_more(self):
        """
        For a dynamic varobj, this indicates whether there appear to be any
        children available. For a non-dynamic varobj, this will be 0.
        """
        try:
            next(self.children())
            return 1
        except StopIteration:
            return 0

    @property
    def dynamic(self):
        """
        This attribute will be present and have the value ‘1’ if the varobj is a
        dynamic varobj. If the varobj is not a dynamic varobj, then this
        attribute will not be present.
        """
        return 1

    @property
    def displayhint(self):
        """
        A dynamic varobj can supply a display hint to the front end. The value
        comes directly from the Python pretty-printer object’s display_hint
        method. See Pretty Printing API.
        """
        return None

    @property
    def frozen(self):
        """
        If the variable object is frozen, this variable will be present with a
        value of 1.
        """
        return 0

    @property
    def exp(self):
        """
        The expression to be shown to the user by the front end to designate
        this child. For example this may be the name of a structure member.
        """
        return self._expr

@micommand("-vdb-var-create")
def _var_create(argv):
    argv_parser = argparse.ArgumentParser()
    argv_parser.add_argument("name")
    argv_parser.add_argument("frame")
    argv_parser.add_argument("expr")
    args = argv_parser.parse_args(argv)

    if args.frame != "*":
        raise Exception("frame designator must be '*'")
    if args.name == "-":
        args.name = _VarObjs.generate_name()
    _VarObjs.validate_name(args.name)

    varobj = _VarObj(args.name, args.expr, parse_and_eval(args.expr))
    _VarObjs.register(varobj)
    return varobj.to_mi()


@micommand("-vdb-var-delete")
def _var_delete(argv):
    argv_parser = argparse.ArgumentParser()
    argv_parser.add_argument("name")
    argv_parser.add_argument("-c", dest="children_only", action="store_true")
    args = argv_parser.parse_args(argv)

    if args.children_only:
        _VarObjs.unregister_children(_VarObjs[args.name])
    else:
        _VarObjs.unregister(_VarObjs[args.name])
    return None


@micommand("-vdb-var-update")
def _var_update(argv):
    # Not yet supported
    return []


@micommand("-vdb-var-duplicate")
def _var_duplicate(argv):
    argv_parser = argparse.ArgumentParser()
    argv_parser.add_argument("name")
    args = argv_parser.parse_args(argv)

    orig = _VarObjs[args.name]
    dup = copy.copy(orig)
    dup._name = _VarObjs.generate_name()
    _VarObjs.register(dup)
    return dup.to_mi()


@micommand("-vdb-var-list-children")
def _var_list_children(argv):
    argv_parser = argparse.ArgumentParser()
    argv_parser.add_argument("--all-values", dest="values", action="store_true")
    argv_parser.add_argument("--no-values", dest="values", action="store_false")
    argv_parser.add_argument("name")
    args = argv_parser.parse_args(argv)

    varobj = _VarObjs[args.name]
    children = list(varobj.children())
    for child in children:
        if not child.name in _VarObjs:
            _VarObjs.register(child)
    return map(_VarObj.to_mi, children)


@micommand("-vdb-var-info-path-expression")
def _var_info_path_expression(argv):
    argv_parser = argparse.ArgumentParser()
    argv_parser.add_argument("name")
    args = argv_parser.parse_args(argv)

    varobj = _VarObjs[args.name]
    if varobj.parent != None:
        raise Error(
            "-vdb-var-info-path-expression is not supported for children of dynamic varobjs"
        )
    else:
        return varobj.exp


@micommand("-vdb-var-set-visualizer")
def _var_set_visualizer(argv):
    argv_parser = argparse.ArgumentParser()
    argv_parser.add_argument("name")
    argv_parser.add_argument("visualizer")
    args = argv_parser.parse_args(argv)

    varobj = _VarObjs[args.name]
    # Now, create visualizer )pretty-printer) object. This is
    # done by evaluating given visualizer expression and and
    # "calling" the result with gdb.Value as parameter).
    #
    # We have to be careful here: the expression is user-supplied
    # and might use stuff from __main__ module. So first try as it is
    # and if if fails, try to evaluate again in context of __main__.
    #
    visualizer_factory = None
    try:
        visualizer_factory = eval(args.visualizer)
    except:
        import __main__

        visualizer_factory = eval(args.visualizer, __main__.__dict__, __main__.__dict__)

    visualizer = visualizer_factory(varobj._value)

    varobj.set_visualizer(visualizer)
