#
# jv:vdb - Visual / VM Debugger
# Copyright (C) 2015, 2021 Jan Vrany
#
# This software is licensed under 'Creative Commons Attribution-NonCommercial 4.0 International License'
#
# You may find a full license text in LICENSE.txt or at http://creativecommons.org/licenses/by-nc/4.0/
#

"""
Pretty printing utilities
"""
import re

import gdb
import gdb.printing

from itertools import chain

def basetypes(ty):
    """
    Return a list of all base types of `ty` (including `ty`!)
    up to the root(s) of the class hierarchy. (Indirect) base
    types are listed before derived types. Handles multiple
    inheritance.
    """
    if ty.code != gdb.TYPE_CODE_STRUCT:
        return []

    def __basetypes(ty, traversed):
        queue = [ ty ]
        basetypes = []
        while len(queue) > 0:
            t = queue.pop()
            if t not in traversed:
                basetypes.append(t)
                traversed.append(t)
                for f in reversed(t.fields()):
                    if f.is_base_class:
                        queue.append(f.type)
                    base = f.type
        return reversed(basetypes)

    return __basetypes(ty, [])



class CxxPrettyPrinter():
    """
    A custom pretty printer for (almost any) C++ object.

    It doesn't do any special printing except it does "flatten"
    members into one flat list (just like in Smalltalk inspector)
    """

    def __init__(self, val):
        self._type = val.dynamic_type
        self._val = val.cast(self._type)

    def to_string(self):
        if self._val.type.code == gdb.TYPE_CODE_PTR:
            return "(%s) 0x%x" % ( self._type, int(self._val) )
        else:
            return "(%s) ... " % ( self._type )

    def fields_of(self, ty):
        return filter(lambda f : f.is_base_class == False and f.artificial == False, ty.fields())

    def fields(self):
        ty = self._type
        if ty.code == gdb.TYPE_CODE_PTR:
            ty = ty.target()
        return chain.from_iterable( (self.fields_of(t) for t in basetypes(ty)) )

    def children(self):

        """
        GDB will call this method on a pretty-printer to compute the children of
        the pretty-printer’s value.

        https://sourceware.org/gdb/onlinedocs/gdb/Pretty-Printing-API.html#Pretty-Printing-API
        """
        val = self._val
        if val.type.code == gdb.TYPE_CODE_PTR:
            val = val.dereference()

        return map(lambda f : (f.name , val[f.name] ), self.fields())

class CxxPtrPrettyPrinter():
    """
    A custom pretty printer for printing pointers (to C++ types)
    """
    def __init__(self, val, contents_visualizer):
        assert val.type.code == gdb.TYPE_CODE_PTR
        self._val = val
        self.contents_visualizer = contents_visualizer

    def to_string(self):
        return self.contents_visualizer(self._val).to_string()


class CxxNullptrPrettyPrinter():
    """
    A custom pretty printer for NULL / nullptr values.
    """
    def __init__(self, val):
        assert val.type.code == gdb.TYPE_CODE_PTR
        assert int(val) == 0x0 # C++ defines nullptr having value of 0
        self._val = val

    def to_string(self):
        return "nullptr"




class CxxCollectionPrettyPrinter(gdb.printing.PrettyPrinter):
    """
    Class implementing a collection of (sub)pretty-printers specialized
    for use with C++ class types.

    It behaves much like GDB's RegexpCollectionPrettyPrinter except that it
    takes C++ inheritance into the account. if there's no pretty-printer
    registered for given (dynamic) type, it looks if there's a pretty-printer
    for supertype(s).

    CxxCollectionPrettyPrinter may nest.
    """

    class CxxSubPrinter(gdb.printing.SubPrettyPrinter):
        def __init__(self, name, regexp, printer):
            super().__init__(name)
            self.regexp = regexp
            self.regexp_c = re.compile(regexp)
            self.printer = printer

        def _lookup(self, t):
            """
            Return self if self is suitable for given type `t`, `None` otherwise.
            """
            if t.code == gdb.TYPE_CODE_PTR:
                t = t.target()
            try:
                if t.name != None and self.regexp_c.search(t.name):
                    return self
            except:
                pass
            return None

        def __call__(self, val):
            """
            Instantiate the pretty-printer for the provided value.
            """

            # Automagically handle NULL / nullptr values so each and every
            # custom pretty printer does not have to care whether it's val
            # is NULL or not.
            #
            # Note, that C++ defines nullptr having value of 0
            if val.dynamic_type.code == gdb.TYPE_CODE_PTR:
                if int(val) == 0x0:
                    return CxxNullptrPrettyPrinter(val)
                else:
                    return CxxPtrPrettyPrinter(val, self.printer)

            if self.printer is not None:
                return self.printer(val)
            return None

    def add_printer(self, name, regexp, printer):
        """Add a printer to the list.

        Add printer to the list.

        Arguments:
            name:    name of this pretty printer (for enable/disable)
            regexp:  the regular expression, as a string.
            printer: a callable that given a value returns an object to
                     pretty-print it.

        Returns:
            Nothing.
        """
        if self.subprinters == None:
            self.subprinters = []
        for subprinter in self.subprinters:
            if subprinter.regexp == regexp:
                subprinter.printer = printer
                return
        self.subprinters.append(
            CxxCollectionPrettyPrinter.CxxSubPrinter(name, regexp, printer)
        )

    def _lookup(self, t):
        """
        Lookup a pretty-printer for given type `t`, traversing supertypes if necessary.
        Returns the pretty-printer factory or None if no pretty-printer is found.
        """
        if self.subprinters == None:
            return None
        if t.code == gdb.TYPE_CODE_PTR:
            t = t.target()
        for b in basetypes(t):
            for subprinter in self.subprinters:
                if subprinter.enabled:
                    printer = subprinter._lookup(b)
                    if printer is not None:
                        return printer
        return None

    def __call__(self, val):
        t = val.dynamic_type
        printer = self._lookup(t)
        if printer is not None:
            return printer(val)

        # Automagically handle NULL / nullptr values. This is to print
        # 'nullptr' instead of '0x0' so it's clear at first glance the
        # pointer is null. Maybe not such a great idea, though...
        #
        # Note, that C++ defines nullptr having value of 0
        if t.code == gdb.TYPE_CODE_PTR and int(val) == 0x0:
            return CxxNullptrPrettyPrinter(val)

        return None
